/* 
   C utilities
   Michael D. Ernst, mernst@theory.lcs.mit.edu
*/

/* Consider making these, with type.h, into a library; then make LDPATH into 
   /usr/lib /usr/local/lib /mylibdir
   (see man ld.so or man ld for more details)
   Avoid using same names as in system libraries.
 
   Or instead just use -I when compiling (this is what I do).
*/

/* ACTION:  change the trivial ones to macros */

#include "c-type.h"
#include <stdarg.h>
#include <stdio.h>

/* See c-util.h for a table of contents. */


/* ***************************************************************************
   ERRORS & TESTING
   **************** */

/* print an error message (postpending "\n") */
void error(char *fmt, ...)
{
  va_list args;
  
  va_start(args, fmt);
  /* fprintf(stderr, "error: "); */
  vfprintf(stderr, fmt, args);
  fprintf(stderr, "\n");
  va_end(args);
}

/* print an error message (postpending "\n"), then die */
void errdie(char *fmt, ...)
{
  va_list args;
  
  va_start(args, fmt);
  /* fprintf(stderr, "error: "); */
  vfprintf(stderr, fmt, args);
  va_end(args);
  fprintf(stderr, "\n");
  exit(1);
}

/* This isn't quite what I'd like it to be, but it's close. */
/* Call it like so: fscanftest(3, ("%d %d %d", &a, &b, &c)) */
#define fscanftest(args, fscanargs)	\
do { if (args != fscanf fscanargs) \
       errdie("Wrong number of values found in fscan%s.", #fscanargs); \
   } while (0)


/* ***************************************************************************
   STRINGS
   ******* */

/* returns TRUE if the strings compare equal, FALSE otherwise */
Bool strequal(char *s, char *t)
{ return(strcmp(s,t) == 0); }

/* modifies the string by removing trailing spaces.  returns its argument */
char *strwrtrim(char *s)
{ size_t i;
  if (s != NULL)
    for (i=strlen(s)-1; i>=0; i--)
      if (isspace(s[i]))
	s[i] = '\0';
      else
	break;
  return s;
}

/* returns pointer to first nonspace character.  */
char *strwltrim(char *s)
{ if (s != NULL)
    while (isspace(*s))
      s++;
  return s;
}

/* removes both leading and trailing spaces.  returns its argument */
char *strwtrim(char *s)
{ if (s != NULL)
    return strwrtrim(strwltrim(s));
  else
    return s;
}

/* modifies the string by removing all whitespace.  returns its argument */
char *strwremove(char *s)
{ char *t = s;
  char *u = s;
  while ((*t = *s++))
    if (isspace(*t))
      t++;
  return u;
}

Bool strempty(char *s)
{ return (s == NULL || *s == '\0'); }

/* returns first index in s at which t can be found, or -1 if it can't */
int strfind(char *s, char *t)
{ int tlen = strlen(t);
  int lastindex = strlen(s) - tlen;
  int i;

  for (i=0; i<=lastindex; i++, s++)
    if (strncmp(s, t, tlen))
      return i;
  return -1;
}

/* like strfind, but returns a pointer or NULL */
char *strpfind(char *s, char *t)
{ int tlen = strlen(t);

  for (; *s != '\0'; s++)
    if (strncmp(s, t, tlen) == 0)
      return s;
  return NULL;
}


/* ***************************************************************************
   ARITHMETIC
   ********** */

#ifdef GCC

/* for GCC; prevent side effects from occuring more than once */
#define min(X, Y)	\
({ typeof(X) __x = (X), __y = (Y); \
   (__x < __y) ? __x : __y; })

#define max(X, Y)	\
({ typeof(X) __x = (X), __y = (Y); \
   (__x > __y) ? __x : __y; })

#else

/* for standard C */
#define min(X, Y)	\
((X) < (Y) ? (X) : (Y))

#define max(X, Y)	\
((X) > (Y) ? (X) : (Y))

#endif

/* tests whether x < y < z */
#define strictrangetest(x, y, z)	\
((x) < (y) && (y) < (z))

int intcmp_forsort(int *a, int *b)
{ return(*a - *b); }

int charcmp_forsort(char *a, char *b)
{ return(*a - *b); }

int strcmp_forsort(char **a, char **b)
{ return strcmp(*a, *b); }




// /* /*
// ***********************************************************************
// ****
// /*    LIST STRUCTURE
// /*    ************** */
// /* 
// /* /* alternative implementation: 
// /*    2-D array and resize (either every time a car or cdr is done, or by a 
// /*    lot when we run off the end)
// /*    this would be more convenient for appending to the end
// /*    could associate with top level the size of the array, the start index
// /*       (usually 0, but larger if we've set the list to its own cdr), and 
// /*       the fill index.
// /*    This would be good for sorting and (binary) searching.
// /* */
// /* 
// /* Cons *cons(Pointer car, Cons *cdr)
// /* { Cons *newcons = (Cons *) malloc(sizeof(Cons *));
// /*   newcons->car = car;
// /*   newcons->cdr = cdr;
// /*   return newcons;
// /* }
// /* 
// /* /* perhaps get rid of these; they're good abstraction nonetheless, so
// /*    maybe just change them to macros */
// /* Pointer car(Cons *cell)
// /* { return cell->car; }
// /* 
// /* Cons *cdr(Cons *cell)
// /* { return cell->cdr; }
// /* 
// /* void setcar(Cons *cell, Pointer car)
// /* { cell->car = car; }
// /* 
// /* void setcdr(Cons *cell, Pointer cdr)
// /* { cell->cdr = cdr; }
// /*      
// /* 
// /* /* foldl and foldr seem to need type parameterization
// /*    base, for instance, must be typed
// /* */
// /* 
// /* int intfoldl(int (*f)(int, int), int base, Cons *list)
// /* {
// /*   while (list != NULL)
// /*     base = (*f)(base, *((int *) car(list)));
// /*   return base;
// /* }
// /* 
// /* int intfoldr(int (*f)(int, int), int base, Cons *list)
// /* {
// /*   if (list == NULL)
// /*     return base;
// /*   else
// /*     return (*f)((int) car(list), intfoldr(f, base, cdr(list)));
// /* }
// /* 
// /* /* iterative implementation */
// /* Cons *imapcar(Pointer (*f)(Pointer), Cons *list)
// /* { 
// /*   Cons *resultcons, *thiscons;
// /* 
// /*   if (list == NULL)
// /*     return NULL;
// /*   thiscons = resultcons = cons((*f)(car(list)), NULL);
// /*   while ((list = cdr(list)) != NULL)
// /*     { setcdr(thiscons, cons((*f)(car(list)), NULL));
// /*       thiscons = cdr(thiscons);
// /*     }
// /*   return resultcons;
// /* }
// /* 
// /* 
// /* /* recursive implementation */
// /* Cons *rmapcar(Pointer (*f)(Pointer), Cons *list)
// /* { 
// /*   if (list == NULL)
// /*     return NULL;
// /*   else
// /*     return cons((*f)(car(list)), rmapcar(f, cdr(list)));
// /* }
// /* 
// /* /* true if the argument is one of the members of the list */
// /* Bool memq(Pointer item, Cons *list)
// /* {
// /*   while (list != NULL)
// /*     if (car(list) == item)
// /*       return TRUE;
// /*   return FALSE;
// /* }
// /* 
// /* /* I hope this name doesn't collide with anything else */
// /* 
// /* /* item will be the first argument to (*comp) */
// /* Bool member(Pointer item, Cons *list, int (*comp)(Pointer, Pointer))
// /* {
// /*   while (list != NULL)
// /*     if ((*comp)(item, car(list)) == 0)
// /*       return TRUE;
// /*   return FALSE;
// /* }
// */
// 
